@@ -66,15 +66,33 @@ register_openers()
 # Base Domain
 BIGML_DOMAIN = os.environ.get('BIGML_DOMAIN', 'bigml.io')
 
+# Domain for prediction server
+BIGML_PREDICTION_DOMAIN = os.environ.get('BIGML_PREDICTION_DOMAIN',
+                                         BIGML_DOMAIN)
+# Whether a prediction server is being used or not
+PREDICTION_SERVER_ON = (BIGML_DOMAIN != BIGML_PREDICTION_DOMAIN)
+
+# Protocol for prediction server
+BIGML_PREDICTION_PROTOCOL = os.environ.get('BIGML_PREDICTION_PROTOCOL',
+                                           'https')
+
 # Check BigML.io host’s SSL certificate
 # DO NOT CHANGE IT.
 VERIFY = (BIGML_DOMAIN == "bigml.io")
 
+# Check prediction server's SSL certificate
+# DO NOT CHANGE IT.
+VERIFY_PREDICTION_SERVER = (BIGML_PREDICTION_DOMAIN == "bigml.com")
+
 # Base URL
 BIGML_URL = 'https://%s/andromeda/' % BIGML_DOMAIN
 # Development Mode URL
 BIGML_DEV_URL = 'https://%s/dev/andromeda/' % BIGML_DOMAIN
 
+# Prediction URL
+BIGML_PREDICTION_URL = '%s://%s/andromeda/' % (BIGML_PREDICTION_PROTOCOL,
+                                               BIGML_PREDICTION_DOMAIN)
+
 # Basic resources
 SOURCE_PATH = 'source'
 DATASET_PATH = 'dataset'
@@ -87,7 +105,8 @@ ENSEMBLE_PATH = 'ensemble'
 ID_PATTERN = '[a-f0-9]{24}'
 SOURCE_RE = re.compile(r'^%s/%s$' % (SOURCE_PATH, ID_PATTERN))
 DATASET_RE = re.compile(r'^(public/)?%s/%s$' % (DATASET_PATH, ID_PATTERN))
-MODEL_RE = re.compile(r'^(public/)?%s/%s$' % (MODEL_PATH, ID_PATTERN))
+MODEL_RE = re.compile(r'^(public/)?%s/%s$|^shared/model/[a-zA-Z0-9]{27}$' % (
+    MODEL_PATH, ID_PATTERN))
 PREDICTION_RE = re.compile(r'^%s/%s$' % (PREDICTION_PATH, ID_PATTERN))
 EVALUATION_RE = re.compile(r'^%s/%s$' % (EVALUATION_PATH, ID_PATTERN))
 ENSEMBLE_RE = re.compile(r'^%s/%s$' % (ENSEMBLE_PATH, ID_PATTERN))
@@ -414,14 +433,16 @@ class BigML(object):
 
         if dev_mode:
             self.url = BIGML_DEV_URL
+            self.prediction_url = BIGML_DEV_URL
         else:
             self.url = BIGML_URL
+            self.prediction_url = BIGML_PREDICTION_URL
 
         # Base Resource URLs
         self.source_url = self.url + SOURCE_PATH
         self.dataset_url = self.url + DATASET_PATH
         self.model_url = self.url + MODEL_PATH
-        self.prediction_url = self.url + PREDICTION_PATH
+        self.prediction_url = self.prediction_url + PREDICTION_PATH
         self.evaluation_url = self.url + EVALUATION_PATH
         self.ensemble_url = self.url + ENSEMBLE_PATH
 
@@ -429,7 +450,7 @@ class BigML(object):
             locale.setlocale(locale.LC_ALL, DEFAULT_LOCALE)
         self.storage = assign_dir(storage)
 
-    def _create(self, url, body):
+    def _create(self, url, body, verify=VERIFY):
         """Creates a new remote resource.
 
         Posts `body` in JSON to `url` to create a new remote resource.
@@ -450,42 +471,53 @@ class BigML(object):
             "status": {
                 "code": code,
                 "message": "The resource couldn't be created"}}
-        try:
-            response = requests.post(url + self.auth,
-                                     headers=SEND_JSON,
-                                     data=body, verify=VERIFY)
 
-            code = response.status_code
+        # If a prediction server is in use, the first prediction request might
+        # return a HTTP_ACCEPTED (202) while the model or ensemble is being
+        # downloaded.
+        code = HTTP_ACCEPTED
+        while code == HTTP_ACCEPTED:
+            try:
+                response = requests.post(url + self.auth,
+                                         headers=SEND_JSON,
+                                         data=body, verify=verify)
+                code = response.status_code
 
-            if code == HTTP_CREATED:
-                location = response.headers['location']
-                resource = json.loads(response.content, 'utf-8')
-                resource_id = resource['resource']
-                error = None
-            elif code in [HTTP_BAD_REQUEST,
-                          HTTP_UNAUTHORIZED,
-                          HTTP_PAYMENT_REQUIRED,
-                          HTTP_FORBIDDEN,
-                          HTTP_NOT_FOUND]:
-                error = json.loads(response.content, 'utf-8')
-                LOGGER.error(error_message(error, method='create'))
-            else:
-                LOGGER.error("Unexpected error (%s)" % code)
+                if code in [HTTP_CREATED, HTTP_OK]:
+                    if 'location' in response.headers:
+                        location = response.headers['location']
+                    resource = json.loads(response.content, 'utf-8')
+                    resource_id = resource['resource']
+                    error = None
+                elif code in [HTTP_BAD_REQUEST,
+                              HTTP_UNAUTHORIZED,
+                              HTTP_PAYMENT_REQUIRED,
+                              HTTP_FORBIDDEN,
+                              HTTP_NOT_FOUND]:
+                    error = json.loads(response.content, 'utf-8')
+                    LOGGER.error(error_message(error, method='create'))
+                elif code != HTTP_ACCEPTED:
+                    LOGGER.error("Unexpected error (%s)" % code)
+                    code = HTTP_INTERNAL_SERVER_ERROR
+
+            except ValueError:
+                LOGGER.error("Malformed response")
+                code = HTTP_INTERNAL_SERVER_ERROR
+            except requests.ConnectionError:
+                LOGGER.error("Connection error")
+                code = HTTP_INTERNAL_SERVER_ERROR
+            except requests.Timeout:
+                LOGGER.error("Request timed out")
+                code = HTTP_INTERNAL_SERVER_ERROR
+            except requests.RequestException:
+                LOGGER.error("Ambiguous exception occurred")
                 code = HTTP_INTERNAL_SERVER_ERROR
-
-        except ValueError:
-            LOGGER.error("Malformed response")
-        except requests.ConnectionError:
-            LOGGER.error("Connection error")
-        except requests.Timeout:
-            LOGGER.error("Request timed out")
-        except requests.RequestException:
-            LOGGER.error("Ambiguous exception occurred")
 
         return maybe_save(resource_id, self.storage, code,
                           location, resource, error)
 
-    def _get(self, url, query_string=''):
+    def _get(self, url, query_string='',
+             shared_username=None, shared_api_key=None):
         """Retrieves a remote resource.
 
         Uses HTTP GET to retrieve a BigML `url`.
@@ -506,7 +538,9 @@ class BigML(object):
             "status": {
                 "code": HTTP_INTERNAL_SERVER_ERROR,
                 "message": "The resource couldn't be retrieved"}}
-
+        auth = (self.auth if shared_username is None
+                else "?username=%s;api_key=%s" % (
+                    shared_username, shared_api_key))
         try:
             response = requests.get(url + self.auth + query_string,
                                     headers=ACCEPT_JSON,
@@ -718,7 +752,7 @@ class BigML(object):
 
         resource = self._get("%s%s" % (self.url, resource_id))
         if resource['code'] == HTTP_OK:
-            if  MODEL_RE.match(resource_id):
+            if MODEL_RE.match(resource_id):
                 return resource['object']['model']['model_fields']
             else:
                 return resource['object']['fields']
@@ -1183,7 +1217,8 @@ class BigML(object):
             body = json.dumps(args)
             return self._create(self.model_url, body)
 
-    def get_model(self, model, query_string=''):
+    def get_model(self, model, query_string='',
+                  shared_username=None, shared_api_key=None):
         """Retrieves a model.
 
            The model parameter should be a string containing the
@@ -1192,17 +1227,22 @@ class BigML(object):
            until it reaches the FINISHED or FAULTY state, the function will
            return a dict that encloses the model values and state info
            available at the time it is called.
+
+           If this is a shared model, the username and sharing api key must
+           also be provided.
         """
         model_id = get_model_id(model)
         if model_id:
             return self._get("%s%s" % (self.url, model_id),
-                             query_string=query_string)
+                             query_string=query_string,
+                             shared_username=shared_username,
+                             shared_api_key=shared_api_key)
 
-    def model_is_ready(self, model):
+    def model_is_ready(self, model, **kwargs):
         """Checks whether a model's status is FINISHED.
 
         """
-        resource = self.get_model(model)
+        resource = self.get_model(model, **kwargs)
         return (resource['code'] == HTTP_OK and
                 get_status(resource)['code'] == FINISHED)
 
@@ -1298,7 +1338,8 @@ class BigML(object):
                     "ensemble": ensemble_id})
 
             body = json.dumps(args)
-            return self._create(self.prediction_url, body)
+            return self._create(self.prediction_url, body,
+                                verify=VERIFY_PREDICTION_SERVER)
 
     def get_prediction(self, prediction):
         """Retrieves a prediction.
